from .rqmodel import RQModel


class RQBool(RQModel):
    _bool: bool

    def __init__(self, state):
        """
        :param state:
        """
        super().__init__()
        self._bool = state

    def get(self):
        """
        get method
        :return:
        """
        return self._bool

    def set(self, value):
        """
        set method
        :param value:
        :return:
        """
        self._bool = value
        self._rq_data_changed.emit()

    def toggle(self):
        """
        toggle method
        :return:
        """
        self.set(not self.get())

    def __bool__(self):
        """
        bool value of rqbool
        :return:
        """
        return self._bool



from PyQt5.QtWidgets import *
from pyreaqtive import *
import sys
import random

from typing import Type


class Counter(RQModel):
    mycounter: RQInt
    rq_list_index: RQInt

    def __init__(self):
        super().__init__()

        self.mycounter = RQInt(10)
        self.rq_list_index = RQInt(0)


class CounterWidget(RQWidget):
    model: Counter

    def __init__(self, model: Counter):
        self.model = model
        super().__init__()

        self.main_layout = QVBoxLayout(self)

        self.counter_increment_button = QPushButton("+")
        self.counter_increment_button.clicked.connect(self.increment)
        self.main_layout.addWidget(self.counter_increment_button)
        self.counter_decrement_button = QPushButton("-")
        self.counter_decrement_button.clicked.connect(self.decrement)
        self.main_layout.addWidget(self.counter_decrement_button)

        self.counter_spin_box = RQSpinBox(self.model.mycounter)
        self.main_layout.addWidget(self.counter_spin_box)

        self.counter_display_label = RQLabel(
            RQFormatter(
                "count x2 {countx2}, index: {index}",
                countx2=RQFunction(
                    lambda count: count * 2,
                    count=self.model.mycounter
                ),
                index=self.model.rq_list_index
            )
        )
        self.main_layout.addWidget(self.counter_display_label)

        self.model.mycounter._rq_data_changed.connect(self.update_counter_text_size)

        self.update_counter_text_size()

    # example of a custom function binding
    def update_counter_text_size(self):
        f = self.counter_display_label.font()
        f.setPointSize(self.model.mycounter)
        self.counter_display_label.setFont(f)

    def increment(self):
        self.model.mycounter.increment()

    def decrement(self):
        self.model.mycounter.decrement()


class AlternateCounterWidget(RQWidget):

    def __init__(self, model: Counter):
        super().__init__()
        self.main_layout = QVBoxLayout(self)
        self.main_layout.addWidget(
            QPushButton("Alternate Counter Widget")
        )

        self.main_layout.addWidget(
            RQLabel(
                RQFormatter(
                    "index is {index}",
                    index=model.rq_list_index
                )
            )
        )


class ChoiceOption(RQModel):
    name: str

    def __str__(self):
        return self.name


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        self.counters = RQList([Counter])
        self.bool_state = RQBool(True)
        self.text_1 = RQText("hello!")

        self.choice_list = RQList()
        self.choice_list.append(ChoiceOption).name = "option 1"
        self.choice_list.append(ChoiceOption).name = "option 2"
        self.choice_list.append(ChoiceOption).name = "option 3"
        self.choice = RQChoice(self.choice_list)

        self.progress = RQInt(0)

        self.some_float = RQFloat(12.7)

        self.init_ui()

    def counter_widget_callback(self, model: Type[RQModel]) -> Type[QWidget]:
        if random.choice([True, False]):
            return CounterWidget(model)
        else:
            return AlternateCounterWidget(model)

    def init_ui(self):
        self.main_widget = QWidget()
        self.setCentralWidget(self.main_widget)
        self.main_layout = QVBoxLayout(self.main_widget)

        self.counters_layout = RQVBoxLayout(model=self.counters, widget_callback=self.counter_widget_callback)
        self.main_layout.addLayout(self.counters_layout)

        self.add_counter = QPushButton("Add Counter")
        self.add_counter.clicked.connect(lambda: self.counters.append(Counter))
        self.main_layout.addWidget(self.add_counter)
        self.remove_counter = QPushButton("Remove Counter")
        self.remove_counter.clicked.connect(lambda: self.counters.pop())
        self.main_layout.addWidget(self.remove_counter)

        self.checkbox_1 = RQCheckbox(self.bool_state, "Checkbox 1")
        self.main_layout.addWidget(self.checkbox_1)
        self.checkbox_2 = RQCheckbox(self.bool_state, "Checkbox 2")
        self.main_layout.addWidget(self.checkbox_2)

        self.line_edit = RQLineEdit(self.text_1)
        self.main_layout.addWidget(self.line_edit)
        self.label_text_1 = RQLabel(RQFormatter("text {text_1}", text_1=self.text_1))
        self.main_layout.addWidget(self.label_text_1)

        self.combobox_1 = RQCombobox(self.choice)
        self.main_layout.addWidget(self.combobox_1)
        self.combobox_2 = RQCombobox(self.choice)
        self.main_layout.addWidget(self.combobox_2)

        self.progress_spin = RQSpinBox(self.progress)
        self.main_layout.addWidget(self.progress_spin)
        self.progress_bar = RQProgressBar(self.progress)
        self.main_layout.addWidget(self.progress_bar)
        self.progress_slider = RQSlider(self.progress)
        self.main_layout.addWidget(self.progress_slider)
        self.progress_dial = RQDial(self.progress)
        self.main_layout.addWidget(self.progress_dial)

        self.some_float_spinbox = RQDoubleSpinBox(self.some_float)
        self.main_layout.addWidget(self.some_float_spinbox)
        self.some_float_display = RQLabel(self.some_float)
        self.main_layout.addWidget(self.some_float_display)

        self.show()


app = QApplication(sys.argv)
mainWindow = MainWindow()

sys.exit(app.exec_())



from .rqmodel import RQModel


class RQFloat(RQModel):
    _float: float

    def __init__(self, initial_float: float):
        super().__init__()
        self._float = initial_float

    def get(self):
        return self._float

    def set(self, value):
        self._float = value
        self._rq_data_changed.emit()

    def __str__(self):
        return str(self._float)

    def __int__(self):
        return int(self._float)

    def __float__(self):
        return self._float



from PyQt5.QtWidgets import QLabel
from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot

from ..models import RQModel

class RQLabel(QLabel):

    model: RQModel = None

    def __init__(self, model, *args):
        super().__init__(str(self.model), *args)
        self.model = model
        self.model._rq_data_changed.connect(self._rq_data_changed)
        self._rq_data_changed()

    def _rq_data_changed(self):
        self.setText(str(self.model))



from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot

from ..models import RQModel


class RQLineEdit(QLineEdit):
    model: RQModel

    def __init__(self, model, *args):
        self.model = model
        super().__init__(*args)
        self.model._rq_data_changed.connect(self._rq_data_changed)
        self.textChanged.connect(self._valueChanged)
        self._rq_data_changed()

    @pyqtSlot()
    def _rq_data_changed(self):
        if not self._rq_self_changing:
            self.setText(str(self.model))

    _rq_self_changing = False

    @pyqtSlot(str)
    def _valueChanged(self, text):
        self._rq_self_changing = True
        self.model.set(text)
        self._rq_self_changing = False



from .rqmodel import RQModel
from .rqlist import RQList

from typing import List, Union


class RQChoice(RQModel):
    _choices: RQList
    _selected: Union[RQModel, None]

    def __init__(self, choices: RQList, selected: RQModel = None):
        super().__init__()
        self._choices = choices
        self._selected = selected
        self.validate_selected()

    def get(self):
        return self._selected

    def get_choices(self):
        return self._choices

    def validate_selected(self):
        if self._selected is None:
            return
        if self._selected not in self._choices:
            raise KeyError

    def set(self, value):
        self._selected = value
        self.validate_selected()
        self._rq_data_changed.emit()

    def __str__(self):
        return self._selected



from PyQt5.QtWidgets import QComboBox
from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot

from ..models import RQChoice


class RQCombobox(QComboBox):
    model: RQChoice

    def __init__(self, model, *args):
        super().__init__(*args)
        self.model = model
        self.model._rq_data_changed.connect(self._rq_data_changed)
        self.model._choices._rq_list_insert.connect(self._rq_choice_insert)
        self.model._choices._rq_list_remove.connect(self._rq_choice_remove)

        for index, choice in enumerate(self.model._choices):
            self._rq_choice_insert(index)

        self.currentIndexChanged.connect(self._currentIndexChanged)

    @pyqtSlot(int)
    def _rq_choice_insert(self, index):
        self.insertItem(
            index,
            str(self.model._choices.get_item(index))
        )

    @pyqtSlot(int)
    def _rq_choice_remove(self, index):
        self.removeItem(index)

    @pyqtSlot()
    def _rq_data_changed(self):
        if not self._rq_self_changing:
            self.setCurrentIndex(
                self.model._choices.get_index(
                    self.model._selected
                )
            )
            pass

    _rq_self_changing = False

    @pyqtSlot(int)
    def _currentIndexChanged(self, index):
        self._rq_self_changing = True
        self.model.set(
            self.model._choices.get_item(index)
        )
        self._rq_self_changing = False



from PyQt5.QtWidgets import QCheckBox
from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot

from ..models import RQModel


class RQCheckbox(QCheckBox):
    model: RQModel = None

    def __init__(self, model, *args):
        super().__init__(*args)
        self.model = model
        self.toggled.connect(self._toggled)
        self.model._rq_data_changed.connect(self._rq_data_changed)
        self._rq_data_changed()

    @pyqtSlot()
    def _rq_data_changed(self):
        if not self._rq_self_changing:
            self.setChecked(bool(self.model))

    _rq_self_changing = False

    @pyqtSlot()
    def _toggled(self):
        self._rq_self_changing = True
        self.model.set(bool(self.checkState()))
        self._rq_self_changing = False



from .rqmodel import RQModel


class RQFormatter(RQModel):
    def __init__(self, format_string, **kwargs):
        super().__init__()
        self.format_string = format_string
        self.variables = kwargs
        for name, model in self.variables.items():
            if isinstance(model, RQModel) or issubclass(type(model), RQModel):
                model._rq_data_changed.connect(self._variable_changed)

    def _variable_changed(self):
        self._rq_data_changed.emit()

    def __str__(self):
        return self.format_string.format(
            **{key: str(variable) for key, variable in self.variables.items()}
        )



from PyQt5.QtWidgets import QVBoxLayout, QWidget, QPushButton
from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot

from ..models import RQModel, RQList

from typing import List


class RQWidget(QWidget):
    """model: RQModel

    def _rq_widget_init(self):
        self.model._rq_data_changed.emit()"""

    pass



from .rqmodel import RQModel
from .rqint import RQInt

from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot

from typing import List, Type


class RQList(RQModel):
    _list: list

    def __init__(self, initial_models: List[Type[RQModel]] = []):
        self._list = []
        for model in initial_models:
            instance = model()
            self._list.append(instance)
        self._update_child_indexes()
        super().__init__()

    _rq_list_insert = pyqtSignal(int)
    _rq_list_remove = pyqtSignal(int)

    def append(self, model: Type[RQModel]) -> RQModel:
        instance = model()
        self._list.append(instance)
        self._update_child_indexes()
        self._rq_list_insert.emit(len(self._list) - 1)
        return instance

    def pop(self):
        if len(self._list) > 0:
            self._list.pop()
            self._update_child_indexes()
            self._rq_list_remove.emit(len(self._list))

    def get_item(self, index):
        return self._list[index]

    def get_index(self, item):
        return self._list.index(item)

    def _update_child_indexes(self):
        for index, item in enumerate(self._list):
            if hasattr(item, "rq_list_index"):
                if isinstance(item.rq_list_index, RQInt):
                    item.rq_list_index.set(index)

    def __iter__(self):
        for item in self._list:
            yield item



from PyQt5.QtWidgets import QProgressBar

from ..models import RQInt


class RQProgressBar(QProgressBar):
    model: RQInt

    def __init__(self, model: RQInt, *args):
        self.model = model
        super().__init__(*args)

        self.model._rq_data_changed.connect(self._rq_data_changed)
        self._rq_data_changed()

    def _rq_data_changed(self):
        self.setValue(self.model.get())



from .rqmodel import RQModel


class RQText(RQModel):
    _text: str

    def __init__(self, text):
        super().__init__()
        self._text = text

    def get(self):
        return self._text

    def set(self, value):
        self._text = value
        self._rq_data_changed.emit()

    def __str__(self):
        return self._text



from .rqmodel import *
from .rqint import *
from .rqbool import *
from .rqtext import *
from .rqlist import *
from .rqchoice import *
from .rqformatter import *
from .rqfunction import *
from .rqfloat import *



from PyQt5.QtWidgets import QVBoxLayout, QWidget, QPushButton
from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot

from ..models import RQModel, RQList
from .rqwidget import RQWidget

from typing import List, Dict, Callable, Type


class RQVBoxLayout(QVBoxLayout):
    model: RQList
    widget_callback: Callable[[Type[RQModel]], Type[QWidget]]

    def __init__(self, model: RQList, widget_callback: Callable[[Type[RQModel]], Type[QWidget]], *args):
        super().__init__(*args)
        self.model = model
        self.widget_callback = widget_callback

        if not isinstance(model, RQList):
            raise TypeError
        else:
            self.model._rq_list_insert.connect(self._rq_insert_widget)
            self.model._rq_list_remove.connect(self._rq_remove_widget)

        for index, item in enumerate(self.model._list):
            self._rq_insert_widget(index)

    widgets: List[QWidget] = []

    @pyqtSlot(int)
    def _rq_insert_widget(self, index):
        model = self.model.get_item(index)
        self.widgets.insert(index, self.widget_callback(model))
        self.addWidget(self.widgets[index])

    @pyqtSlot(int)
    def _rq_remove_widget(self, index):
        self.removeWidget(self.widgets[index])
        self.widgets[index].deleteLater()
        self.widgets.pop(index)



from setuptools import setup

setup(
    name='pyreaqtive',
    version='0.1.0',
    packages=['pyreaqtive', 'pyreaqtive.models', 'pyreaqtive.widgets'],
    url='https://github.com/polfeliu/pyreaqtive',
    license='Attribution-NoDerivatives 4.0 International (CC BY-ND 4.0)',
    author='Pol Feliu Cuberes',
    author_email='feliupol@gmail.com',
    keywords='pyqt5 framework reactive gui',
    description='PyQt Reactive Framework',
    install_requires=[
        "pyqt5>=5.6"
    ]
)



from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot


class RQModel(QObject):
    """
    RQModel Base Class
    All pyqtreactive models must inherit from this class, that provides basic get, set method and data changed signals
    """

    def get(self):
        """
        Method to get the value of the underlying object
        Must be overridden by model
        :return: value of the object
        """
        raise NotImplementedError

    def set(self, value):
        """
        Method to set the value of the underlying object
        Must be overridden by model
        :param value: value to be set to the object
        :return: None
        """
        raise NotImplementedError

    _rq_data_changed = pyqtSignal()



from .rqmodel import RQModel


class RQInt(RQModel):
    _int: int

    def __init__(self, integer):
        super().__init__()
        self._int = integer

    def get(self):
        return self._int

    def set(self, value):
        self._int = value
        self._rq_data_changed.emit()

    def increment(self, delta=1):
        self.set(self._int + delta)

    def decrement(self, delta=1):
        self.increment(delta=-delta)

    def __str__(self):
        return str(self._int)

    def __int__(self):
        return self._int

    def __float__(self):
        return float(self._int)



from .rqwidget import *
from .rqlabel import *
from .rqspinbox import *
from .rqcheckbox import *
from .rqvboxlayout import *
from .rqlineedit import *
from .rqcombobox import *
from .rqprogressbar import *
from .rqslider import *
from .rqdial import *
from .rqdoublespinbox import *



from PyQt5.QtWidgets import QSpinBox
from PyQt5.QtCore import Qt, QObject, pyqtSignal, pyqtSlot

from ..models import RQInt


class RQSpinBox(QSpinBox):
    model: RQInt = None

    def __init__(self, model, *args):
        super().__init__(*args)
        self.model = model
        self._rq_data_changed()
        self.model._rq_data_changed.connect(self._rq_data_changed)
        self.valueChanged.connect(self._valueChanged)

    @pyqtSlot()
    def _rq_data_changed(self):
        if not self._rq_self_changing:
            self.setValue(int(self.model))

    _rq_self_changing = False

    @pyqtSlot()
    def _valueChanged(self):
        self._rq_self_changing = True
        self.model.set(self.value())
        self._rq_self_changing = False
